package com.lanren.huhu.partner.schedule;

import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.lanren.huhu.partner.constants.Constants;
import com.lanren.huhu.partner.domain.*;
import com.lanren.huhu.partner.model.AgentSettle;
import com.lanren.huhu.partner.model.AgentSnapshot;
import com.lanren.huhu.partner.service.*;
import com.lanren.huhu.partner.util.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.datasource.DataSourceTransactionManager;
import org.springframework.stereotype.Component;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;

import static org.springframework.util.StringUtils.isEmpty;

/**
 * @author chen
 * @title: AgentSettleTask
 * @projectName partner
 * @description: 代理商结算
 * @package com.lanren.huhu.partner.schedule
 * @date 2019-07-23 10:31
 */
@Component
public class AgentSettleTask {
    private static Logger logger = LoggerFactory.getLogger(AgentSettleTask.class);
    private static final int PAGE_SIZE = 1000;
    private static final BigDecimal ZERO = new BigDecimal(0);

    @Autowired
    AgentRewardService agentRewardService;
    @Autowired
    OrdersActiveService ordersActiveService;
    @Autowired
    OrderTaobaoJdCommissionService orderTaobaoJdCommissionService;
    @Autowired
    OrderTaobaoJdGoodsCommissionService orderTaobaoJdGoodsCommissionService;
    @Autowired
    DataSourceTransactionManager dataSourceTransactionManager;
    @Autowired
    TransactionDefinition transactionDefinition;
    @Autowired
    AgentAccountService agentAccountService;
    @Autowired
    AgentIncomeSummaryService agentIncomeSummaryService;
    @Autowired
    AgentAccountLogService agentAccountLogService;

    public void runScheduledTask() {
        logger.info("run AgentSettleTask");
        try {
            long input = DateUtils.getLastMonth1stTimestamp(System.currentTimeMillis());
            long beginTs = DateUtils.getMonth1stTimestamp(input) / 1000L;
            long endTs  = DateUtils.getMonthLastTimestamp(input) / 1000L;
            updateAll(beginTs, endTs);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    private void updateAll(long beginTs, long endTs) {
        updateOrderCommission(beginTs, endTs, 0, 0);
        doSettle(beginTs, endTs);
    }

    private void doSettle(long beginTs, long endTs) {
        AgentSettle errAgentSettle = null;
        try {
            int pageNo = 1;
            while (true) {
                Page<AgentSettle> page = new Page<AgentSettle>(pageNo, PAGE_SIZE);
                page.setOptimizeCountSql(false);
                page.setSearchCount(false);
                IPage<AgentSettle> ipage = agentRewardService.selectAgentSettlePageByTs(page, beginTs, endTs);
                List<AgentSettle> agentSettleList = ipage.getRecords();
                if (agentSettleList != null && agentSettleList.size() > 0) {
                    for (AgentSettle agentSettle : agentSettleList) {
                        errAgentSettle = agentSettle;
                        if (!doDbUpdate(agentSettle, beginTs, endTs)) {
                            logger.error("doDbUpdate failed for agent {}", agentSettle.getAgentId());
                        }
                    }
                    pageNo++;
                } else {
                    break;
                }
            }
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            logger.error("errAgentSettle is {}", errAgentSettle);
        }
    }
    /**
     * 自购省和分享赚的奖励金额 需要根据最终的实际原始佣金 重新计算
     * @param beginTs
     * @param endTs
     */
    private void updateOrderCommission(long beginTs, long endTs, int runType, int agentId) {
        AgentReward errAgentReward = null;
        try {
            int pageNo = 1;
            while (true) {
                Page<AgentReward> page = new Page<AgentReward>(pageNo, PAGE_SIZE);
                page.setOptimizeCountSql(false);
                page.setSearchCount(false);
                IPage<AgentReward> ipage;
                if (runType == 0) {
                    ipage = agentRewardService.selectOrderRewardPageByTs(page, beginTs, endTs);
                } else {
                    ipage = agentRewardService.selectOrderRewardPageByTsAndAgentId(page, beginTs, endTs, agentId);
                }
                List<AgentReward> agentRewardList = ipage.getRecords();
                if (agentRewardList != null && agentRewardList.size() > 0) {
                    /**
                     * 循环处理每条奖励
                     */
                    for (AgentReward agentReward : agentRewardList) {
                        errAgentReward = agentReward;
                        /**
                         * 自购省奖励: reward_type: 300 或 750, settle_state > 0 且 != 300
                         * 需要根据订单号查找最终的真实的原始佣金 并 按比例快照重新计算奖励金额,
                         * 更新字段: all_money_ori cash commission_account
                         * 如果订单失效: 更新settle_state, all_money_ori/cash/commission_account 抹平, 跳过处理, 循环到下一条
                         */
                        if ( (agentReward.getRewardType() == Constants.AGENT_REWARD_TYPE_PURCHASE || agentReward.getRewardType() == Constants.AGENT_REWARD_TYPE_SELF_ORDER)
                                && agentReward.getSettleState() > 0
                                && agentReward.getSettleState() != Constants.SETTLE_STATE_DONE) {
                            if (isEmpty(agentReward.getOrderType()) || isEmpty(agentReward.getOrderSn()) || isEmpty(agentReward.getRatioAll())) {
                                /**
                                 * 历史数据: 没有记录订单号 或者 没有记录比例 不更新奖励金额
                                 */
                            } else {
                                /**
                                 * 有订单号 和 比例快照 可以更新金额
                                 */
                                OrdersActive ordersActive = ordersActiveService.getOneByOrderSn(agentReward.getOrderType(), agentReward.getOrderSn());
                                OrderTaobaoJdCommission order = orderTaobaoJdCommissionService.getOneByOrderSn(agentReward.getOrderType(), agentReward.getOrderSn());
                                if (null == ordersActive || null == order) {
                                    /**
                                     * 如果没有找到active 或者 没有找到commission
                                     * 不更新奖励金额
                                     */
                                } else {
                                    /**
                                     * 订单状态只有三种: -3, 1, 3
                                     */
                                    if (order.getStatus() == Constants.ORDER_STATE_FAILED) {
                                        /**
                                         * 订单失效 订单付款 跳过, 处理下一条奖励
                                         */
                                        updateReward2Failed(agentReward);
                                        continue;
                                    } else if (order.getStatus() == Constants.ORDER_STATE_PAID) {
                                        /**
                                         * 订单还是付款状态 更新奖励时间 下个月再处理
                                         */
                                        long newRechargeTime = DateUtils.getNextMonth1stTimestamp(agentReward.getRechargeTime().getTime());
                                        agentReward.setRechargeTime(new Date(newRechargeTime));
                                        agentRewardService.updateById(agentReward);
                                        continue;
                                    } else if (order.getStatus() == Constants.ORDER_STATE_SETTLED) {
                                        /**
                                         * 订单结算 更新金额
                                         */
                                        if (ordersActive.getIsForNew() == 1) {
                                            /**
                                             * 新人专享订单 不给合伙人奖励 跳过
                                             */
                                            updateReward2Failed(agentReward);
                                            continue;
                                        } else {
                                            /**
                                             * 正常订单 更新原始佣金
                                             * all_money_ori: 原始佣金
                                             * cash: all_money_ori 乘以 ratio_all中的每个比例
                                             * commission_account: cash 乘以 commission_rate
                                             */
                                            updateRewardCommission(agentReward, order.getAllMoney());
                                        }
                                    }
                                }
                            }
                        }
                        /**
                         * 分享赚订单: 按子订单号更新奖励金额 60 或 760
                         */
                        if ( (agentReward.getRewardType() == Constants.AGENT_REWARD_TYPE_SHARE || agentReward.getRewardType() == Constants.AGENT_REWARD_TYPE_SHARE_ORDER)
                                && agentReward.getSettleState() > 0
                                && agentReward.getSettleState() != Constants.SETTLE_STATE_DONE) {
                            if (isEmpty(agentReward.getOrderType()) || isEmpty(agentReward.getOrderSn()) || isEmpty(agentReward.getSubOrderSn())
                                    || isEmpty(agentReward.getRatioAll())) {
                                /**
                                 * 历史数据: 没有记录子订单号 或者 没有记录比例 不更新奖励金额
                                 */
                            } else {
                                /**
                                 * 有订单号 和 比例快照 可以更新金额
                                 */
                                OrderTaobaoJdGoodsCommission subOrder = orderTaobaoJdGoodsCommissionService.getOneByOrderSn(agentReward.getOrderType(), agentReward.getOrderSn(), agentReward.getSubOrderSn());
                                if (null == subOrder) {
                                    /**
                                     * 如果没有找到子订单
                                     * 不更新奖励金额
                                     */
                                } else {
                                    /**
                                     * 订单状态只有三种: -3, 1, 3
                                     */
                                    if (subOrder.getOrderState() == Constants.ORDER_STATE_FAILED) {
                                        /**
                                         * 订单失效 订单付款 跳过, 处理下一条奖励
                                         */
                                        updateReward2Failed(agentReward);
                                    } else if (subOrder.getOrderState() == Constants.ORDER_STATE_PAID) {
                                        /**
                                         * 订单还是付款状态 更新奖励时间 下个月再处理
                                         */
                                        long newRechargeTime = DateUtils.getNextMonth1stTimestamp(agentReward.getRechargeTime().getTime());
                                        agentReward.setRechargeTime(new Date(newRechargeTime));
                                        agentRewardService.updateById(agentReward);
                                        continue;
                                    } else if (subOrder.getOrderState() == Constants.ORDER_STATE_SETTLED) {
                                        /**
                                         * 正常订单 更新原始佣金
                                         * all_money_ori: 原始佣金
                                         * cash: all_money_ori 乘以 ratio_all中的每个比例
                                         * commission_account: cash 乘以 commission_rate
                                         */
                                        updateRewardCommission(agentReward, subOrder.getAllMoney());
                                    }
                                }
                            }
                        }
                    }
                    pageNo++;
                } else {
                    break;
                }
            }
        } catch (Exception e) {
            logger.error(e.getMessage(),e);
            logger.error("errAgentReward is {}", errAgentReward);
        }
    }

    private void updateRewardCommission(AgentReward agentReward, BigDecimal allMoney) {
        agentReward.setAllMoneyOri(allMoney);
        String ratioAllStr = agentReward.getRatioAll().replaceAll("[\\[\\],\"]", "");
        String[] strs = ratioAllStr.split(",");
        BigDecimal cash = allMoney;
        for (String s : strs) {
            BigDecimal ratio = new BigDecimal(s);
            cash = cash.multiply(ratio);
        }
        agentReward.setMoney(cash);
        agentReward.setCommissionAcount(cash);
        BigDecimal commission = cash.multiply(agentReward.getCommissionRate());
        agentReward.setAmount(commission);
        agentReward.setCommission(commission);
        agentReward.setSettleState(Constants.SETTLE_STATE_DONE);
        agentRewardService.updateById(agentReward);
    }

    /**
     * money 面额(message里, 需要上供的代理商自己获得的奖励)
     * commission_account 面额
     * amount 奖励金额
     * commission 奖励金额
     * cash_code 支付金额
     */
    private void updateReward2Failed(AgentReward agentReward) {
        agentReward.setAllMoneyOri(ZERO);
        agentReward.setMoney(ZERO);
        agentReward.setCommissionAcount(ZERO);
        agentReward.setCommission(ZERO);
        agentReward.setCashCode("0");
        agentReward.setSettleState(Constants.SETTLE_STATE_FAILED);
        agentReward.setRewardStatus(Constants.REWARD_STATUS_SETTLE);
        agentRewardService.updateById(agentReward);
    }

    public boolean doOneAgentSettle(Integer agentId) {
        long input = DateUtils.getLastMonth1stTimestamp(System.currentTimeMillis());
        long beginTs = DateUtils.getMonth1stTimestamp(input) / 1000L;
        long endTs  = DateUtils.getMonthLastTimestamp(input) / 1000L;
        AgentSettle errAgentSettle = null;
        try {
            updateOrderCommission(beginTs, endTs, 1, agentId);
            AgentSettle agentSettle = agentRewardService.selectAgentSettleByTsAndAgentId(beginTs, endTs, agentId);
            errAgentSettle = agentSettle;
            return doDbUpdate(agentSettle, beginTs, endTs);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            logger.error("agentId is {}, errAgentSettle is {}", agentId, errAgentSettle);
            return false;
        }
    }

    private boolean doDbUpdate(AgentSettle agentSettle, long beginTs, long endTs) {
        TransactionStatus transactionStatus = null;
        try {
            logger.info("agentSettle is {}", agentSettle);
            /**
             * 开启事务
             */
            transactionStatus = dataSourceTransactionManager.getTransaction(transactionDefinition);

            /**
             * 获取合伙人账户 锁定行
             */
            AgentAccount agentAccount = agentAccountService.getOneByUserIdForUpdate(agentSettle.getAgentId());
            /**
             * 插入变更记录
             */
            insertIntoAgentAccountLog(agentSettle, agentAccount);
            updateAgentAccount(agentSettle, agentAccount);
            updateAgentRewardStatus(agentSettle.getAgentId(), beginTs, endTs);
            insertAgentIncomeSummary(agentSettle, beginTs);
            insertAgentSalary(agentSettle);
            /**
             * 提交事务
             */
            dataSourceTransactionManager.commit(transactionStatus);
//            dataSourceTransactionManager.rollback(transactionStatus);
            return true;
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            /**
             * 回滚事务
             */
            if (transactionStatus != null) {
                logger.info("do settle failed, rollback");
                dataSourceTransactionManager.rollback(transactionStatus);
            }
            return false;
        }
    }

    private void insertAgentSalary(AgentSettle agentSettle) {

    }

    private void insertAgentIncomeSummary(AgentSettle agentSettle, long beginTs) {
        Date now = new Date(beginTs * 1000L);
        int yeartime = Integer.parseInt(DateUtils.getYear(now));
        int monthtime = Integer.parseInt(DateUtils.getMonth(now));
        AgentIncomeSummary agentIncomeSummary = agentIncomeSummaryService.getOneByAgentidYearMonth(agentSettle.getAgentId(), yeartime, monthtime);
        if (null == agentIncomeSummary) {
            agentIncomeSummary = new AgentIncomeSummary();
            agentIncomeSummary.setUserId(agentSettle.getUserId());
            agentIncomeSummary.setAgentId(agentSettle.getAgentId());
            agentIncomeSummary.setYeartime(yeartime);
            agentIncomeSummary.setMonthtime(monthtime);
            agentIncomeSummary.setIncome(ZERO);
            agentIncomeSummary.setContent("");
            agentIncomeSummary.setState(15);
            agentIncomeSummaryService.save(agentIncomeSummary);
        } else {
            agentIncomeSummary.setState(15);
            agentIncomeSummaryService.updateById(agentIncomeSummary);
        }
    }

    private int updateAgentRewardStatus(int agentId, long beginTs, long endTs) {
        return agentRewardService.updateRewardStatusByUidAndTs(agentId, beginTs, endTs);
    }

    private void updateAgentAccount(AgentSettle agentSettle, AgentAccount agentAccount) {
        /**
         * 更新合伙人账户金额
         */
        agentAccount.setBalance(agentAccount.getBalance()
                .add(agentSettle.getBalance())
                .subtract(agentSettle.getTechCharge()));
        agentAccountService.updateById(agentAccount);
    }

    private void insertIntoAgentAccountLog(AgentSettle agentSettle, AgentAccount agentAccount) {
        /**
         * 月结发钱
         */
        AgentSnapshot snapshot = new AgentSnapshot();
        snapshot.setAgentId(agentAccount.getAgentId());
        snapshot.setBalance(agentAccount.getBalance());
        BigDecimal changeNum = agentSettle.getBalance();
        BigDecimal changedAmount = agentAccount.getBalance().add(changeNum);
        AgentAccountLog log = createNewAgentAccountLog(agentSettle,
                Constants.PARTNER_ACCOUNT_LOG_TYPE_SETTLE,
                changeNum,
                changedAmount,
                Constants.PARTNER_ACCOUNT_LOG_TYPE_SETTLE_REMARK,
                snapshot.toString());
        agentAccountLogService.save(log);
        /**
         * 平台技术服务费
         */
        if (agentSettle.getTechCharge().compareTo(ZERO) > 0) {
            changeNum = agentSettle.getTechCharge().negate();
            changedAmount = changedAmount.add(changeNum);
            log = createNewAgentAccountLog(agentSettle,
                    Constants.PARTNER_ACCOUNT_LOG_TYPE_TECH_CHARGE,
                    changeNum,
                    changedAmount,
                    Constants.PARTNER_ACCOUNT_LOG_TYPE_TECH_CHARGE_REMARK,
                    "");
            agentAccountLogService.save(log);
        }
    }

    private AgentAccountLog createNewAgentAccountLog(AgentSettle agentSettle, int type, BigDecimal changeNum, BigDecimal changedAmount, String payRemark, String snapshot) {
        Date now = new Date();
        int yeartime = Integer.parseInt(DateUtils.getYear(now));
        int monthtime = Integer.parseInt(DateUtils.getMonth(now));
        int daytime = Integer.parseInt(DateUtils.getDay(now));
        AgentAccountLog log = new AgentAccountLog();
        log.setUserId(agentSettle.getUserId());
        log.setAgentId(agentSettle.getAgentId());
        log.setType(type);
        log.setChangeNum(changeNum);
        log.setChangedAmount(changedAmount);
        log.setPayRemark(payRemark);
        log.setPayType(false);
        log.setYeartime(yeartime);
        log.setMonthtime(monthtime);
        log.setDaytime(daytime);
        log.setSnapshot(snapshot);
        return log;
    }
}
